/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals ROT, Map, Player, Util, console */
/* globals TurnProcessor, Order, NullOrder, PlayerOrders */
/* globals ShipOrderReport, ShipDesign, ShipDesigner, ShipOrder */
/* exported stylenames */

const InitialState = { version: "37", minimumVersion: "37", turn: 1 }; /* Initial Game State - version indicates which version of save data we create and can read */
const stylenames = ['player', 'alkari', 'auron', 'caffe', 'clanger', 'drahmini', 'gobb', 'iyanden', 'jerandi', 'kalli', 'kelle',
	'pican', 'regul', 'saggitt', 'strata', 'fleet', 'worati'];
const techs = [0, 8, 9, 6, 6, 5, 3, 1, 8, 8, 3, 9, 4, 5, 3, 2, 0];
var Game = {
	STYLENAMES: stylenames,
	FULLNAMES: {'player': "you",
		/*1*/	'alkari': "the Alkari",
		/*2*/	'auron':"the Aurons",
		/*3*/		'caffe':"the Caffenians",
		/*4*/		'clanger':"the Clangers",
		/*5*/		'drahmini':"the Drahmini",
		/*6*/		'gobb':"the Gobbannanean Principalities",
		/*7*/		'iyanden':"the Chroniomancy of Iyanden",
		/*8*/		'jerandi':"the Jerandi People's Republic",
		/*9*/		'kalli':"the Kallikanzaroi",
		/*10*/		'kelle':"the Kellesonians",
		/*11*/		'pican':"the Picans",
		/*12*/		'regul':"the Regul",
		/*13*/		'saggitt':"the Saggittarians",
		/*14*/		'strata':"Strata Inc.",
		/*15*/		'fleet':"the Fleet",
		/*16*/		'worati':"the Worati"
				},
	State: { version: undefined, baseMap: undefined, turn: undefined, playerProduction: undefined },
	_standardShipDesigns: [],
	_players: [],
	_stylenameToPlayerMapping: {},
	_COMPUTERURL: null,
	_computers: {},
	_orders:{},
	History: { baseMap: {} },
	ARTIFACT_DESIGN: null,
	_showMap: function(map) {
		map.forEachHex(function(hex) {
			const hexelement = document.getElementById(hex.id);
			if (hex.lastScanned) {
				const planetElement = hexelement.childNodes[3];
				planetElement.classList.remove("starbase","planet");
				planetElement.classList.add("emptyspace");
				if (hex.starbase) {
					planetElement.classList.remove("emptyspace");
					planetElement.classList.add("starbase");
				}
				if (hex.planet) {
					planetElement.classList.remove("emptyspace");
					planetElement.classList.add("planet");
				}
				hexelement.classList.remove(...stylenames);
				if (hex.lastScanned && hex.owner) {
					hexelement.classList.add(hex.owner);
				}
				let text = "";
				if (hex.level) {
					text = "Class "+hex.level+" planet\n";
				}
				if (hex.owner) {
					text += "Owned by "+Game.FULLNAMES[hex.owner]+"\n";
				}
				hexelement.title = text + "Last report from Turn " + hex.lastScanned;
				if (hex.lastReport) {
					hexelement.title += "\n"+hex.lastReport;
				}
				hexelement.classList.remove("fogofwar");
			} else {
				hexelement.classList.add("fogofwar");
			}
		});
	},
	_generateNewGame: function() {
		Game.State = Object.assign({}, InitialState);
		Game.State.baseMap = new Map();
		Game.History = { baseMap: {}, version: Game.State.version };
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.planet) Game.History.baseMap[hex.id] = true;
		}); /* basemap shows where the planets are */

		// Now create the players
		Game._createPlayers(document.getElementById("favouredTech").selectedIndex);

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);
		Game.State.turn = 0;
		Game._newTurn();
	},
	_createPlayers: function(playerTechIndex) {
		const seed = ROT.RNG.getSeed();
		const capital = {};
		Game.State.baseMap.forEachHex(hex => {
			if (hex.owner) {
				capital[hex.owner] = hex.id;
			}
		});
		Game._players = [];
		Game._stylenameToPlayerMapping = {};
		for(let i=0; i< Game.STYLENAMES.length; i++) {
			let player = new Player(Game.STYLENAMES[i],
				capital[Game.STYLENAMES[i]],
				(i === 0 ? playerTechIndex : techs[i]),
				seed);
			player.init(Game.State.baseMap,Game.State.turn);
			Game._players.push(player);
			Game._stylenameToPlayerMapping[Game.STYLENAMES[i]] = player;
		}
	},
	_newTurn: function() {
		Game.State.turn++;
		Game._updateHistory();
		Game._updateProduction();
		Game._saveGameState();
		Game._startTurn();
	},
	_startTurn: function() {
		Game._pokeComputerWorkers();
		Game._clearDisplayTables();
		Game._showResearchReports();
		Game._showColonyReports();
		ShipOrderReport.showShipReports(Game._players[0]);
		Game._showMap(Game._players[0].map);
		PlayerOrders.prepare();
		Util.replaceText("turnnumber", Game.State.turn);
		Game._orders[Game.STYLENAMES[0]] = null; // No orders from the player
	},
	_createComputerWorkers: function() {
		Game._computers = {};
		Game._orders = {};
		for(let i=1; i<Game.STYLENAMES.length; i++) {
			let worker = new Worker(Game._COMPUTERURL);
			worker.onmessage = Util.captureExceptionEventHandler(Game._computerHasOrders);
			worker.onerror = Util.captureExceptionEventHandler(Game._computerHasError);
			Game._computers[Game.STYLENAMES[i]] = worker;
		}
	},
	_pokeComputerWorkers: function() {
		for(let i=1; i<Game._players.length; i++) {
			let player = Game._players[i];
			let worker = Game._computers[player.stylename];
			Game._orders[player.stylename] = null;
			worker.postMessage({stylename: player.stylename,
								baseMap: JSON.stringify(Game.State.baseMap),
								capital: player.capital,
								tech:player.indexOfFavouredTech,
								player: JSON.stringify(player)
								});
			player.seed = 0; // So that the computer worker knows it's already seen the seed
		}
	},
	_clearDisplayTables: function() {
	// Clear existing rows after the templates
		for(let templaterow of document.querySelectorAll(".template")) {
			while (templaterow.nextSibling) {
				templaterow.parentNode.removeChild(templaterow.nextSibling);
			}
		}
	},
	_updateProduction: function() {
		const productionTotal = {};
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.owner && hex.level > 0) {
				productionTotal[hex.owner] = productionTotal[hex.owner] ? productionTotal[hex.owner]+hex.level*20 : hex.level*20;
			}
		});
		for(let player of Game._players) {
			player.production = productionTotal[player.stylename];
		}
	},
	_showColonyReports: function() {
		// Find the colony listing table
		let colonies = document.getElementById("colonies");
		Game._players[0].map.forEachHex(function(hex) {
			if (hex.owner == Game.STYLENAMES[0] && hex.level > 0) {
				const newrow = colonies.insertRow();
				const rowname = document.createElement("td"); rowname.append(hex.id);
				const report = document.createElement("td");
				report.append(hex.colonyReport ? hex.colonyReport : "Nothing to report");
				const marines = document.createElement("td"); marines.append(hex.marines);
				const level = document.createElement("td"); level.append(hex.level);
				const productionCell = document.createElement("td"); productionCell.append(hex.level * 20);
				newrow.append(rowname,report,marines,level,productionCell);
			}
		});
	},
	_showResearchReports: function() {
		Game._players[0].forEachTech( (tech, research) => {
			Util.replaceText(tech.domPrefix+"Points",research);
			let level = tech.level(research);
			Util.replaceText(tech.domPrefix+"Level",level);
			while(level > 1) {
				document.getElementById(tech.domPrefix+level).classList.remove("hidden");
				level--;
			}
		});
	},
	_updateHistory: function() {
		let sheet = {};
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.owner) sheet[hex.id] = hex.owner; /* Each turn has a map showing who owns the hex */
		});
		Game.History[Game.State.turn] = sheet;
		localStorage.domainOfTheAncientsHistory = JSON.stringify(Game.History);
	},
	init: function() {
		// Find the URL for the computer.js script
		const computerjs = document.getElementById("djv_domainOfTheAncients_computer-js");
		Game._COMPUTERURL = computerjs.src;

		ShipDesign.init();
		for(let shipdesign of ShipDesign.REGISTRY) { // Only valid because the players haven't been created or loaded yet!
			Game._standardShipDesigns.push(shipdesign);
		}
		Game.ARTIFACT_DESIGN = ShipDesign.REGISTRY[0];

		Game._initGameButtons();
		Game._initTwisties();
		PlayerOrders.init();
		Game._createComputerWorkers();
		ShipDesigner.init();
	},
	_initGameButtons: function() {
		document.getElementById("generate").onclick = Util.captureExceptionFunction(Game._generateNewGame);
		if (localStorage.domainOfTheAncients && localStorage.domainOfTheAncientsVersion && localStorage.domainOfTheAncientsSaveDate) {
			Game._showLoadButton();
			let v;
			try {
				v = JSON.parse(localStorage.domainOfTheAncientsVersion);
			} catch(_e) {
				v = null; // Clearly no version data if we can't parse it!
			}
				if (v && v.game && (parseInt(v.game) >= InitialState.minimumVersion)) {
					document.getElementById("resume").disabled = false;
					document.getElementById("wrongVersion").classList.replace("visible","hidden");
					document.getElementById("resume").onclick = Util.captureExceptionFunction(Game._load);
				} else {
					document.getElementById("resume").disabled = true;
					if (v && v.plugin) {
						Util.replaceText("incompatibleVersionNumber",v.plugin);
					}
					document.getElementById("wrongVersion").classList.replace("hidden","visible");
				}
		} else {
			document.getElementById("resume").disabled = true;
		}
		if (localStorage.domainOfTheAncientsHistory) {
			// Nothing yet!
		} else {
			document.getElementById("replay").disabled = true;
		}

		document.getElementById("nextTurn").onclick = Util.captureExceptionFunction(Game._humanHasOrders);
		document.getElementById("processorderread").onclick = Util.captureExceptionFunction(Game._completeTurn);
	},
	_initTwisties: function() {
		// Look for the twisties and set them up
		for(let button of document.querySelectorAll("h4 button")) {
			// Look for the child svg's 'we are interested in
			let svg1 = null, svg2 = null, followingdiv = null;
			for(let child of button.childNodes) {
				if (child.nodeName == "svg") {
					if (svg1 === null) {
						svg1 = child;
					} else {
						svg2 = child;
					}
				}
			}
			// And then the uncle's (i.e. parent's sibling) div
			let sibling = button.parentNode.nextSibling;
			while(followingdiv === null && sibling) {
				if (sibling.nodeName === "DIV") {
					followingdiv = sibling;
				} else {
					sibling = sibling.nextSibling;
				}
			}

			if (svg1 && svg2 && followingdiv) {
				button.onclick = Util.captureExceptionFunction(Util.toggleFields,[svg1,svg2,followingdiv]);
			}
		}
	},
	_computerHasError: function(event) {
		Util.reportException(event);
		const worker = event.currentTarget;
		for(let stylename of Object.keys(Game._computers)) {
			if (Game._computers[stylename] === worker) {
				Game._orders[stylename] = [new NullOrder()];
				console.log(`It was the worker for ${stylename} that encountered ${event}`);
				break;
			}
		}
	},
	_computerHasOrders: function(event) {
		const worker = event.currentTarget;
		for(let stylename of Object.keys(Game._computers)) {
			if (Game._computers[stylename] === worker) {
				const spend = JSON.parse(event.data.orders);
				Game._orders[stylename] = [];
				for(let orderdata of spend) {
					Game._orders[stylename].push(Order.fromJSON(orderdata));
				}
				const shiporders = JSON.parse(event.data.ships);
				for(let i=0; i<shiporders.length; i++) {
					Game._stylenameToPlayerMapping[stylename].ships[i].order = ShipOrder.fromJSON(shiporders[i]);
				}
				Game._stylenameToPlayerMapping[stylename].ROTRNGState = event.data.ROTRNGState; // Ready to resend next time
				break;
			}
		}
		if (!Game._processIfEverybodyReady()) {
			Game._updateAwaitingList();
		}
	},
	_humanHasOrders: function() {
		Game._orders[Game.STYLENAMES[0]] = PlayerOrders.getOrders();
		if (!Game._processIfEverybodyReady()) {
			// Need to wait for more orders from computer players, so display the status page
			document.getElementById("summary").classList.add("hidden");
			document.getElementById("awaitingOrders").classList.remove("hidden");
			Game._updateAwaitingList();
		}
	},
	_updateAwaitingList: function() {
		if (!document.getElementById("awaitingOrders").classList.contains("hidden")) {
			const list = document.getElementById("awaitingOrdersFromList");
			Util.emptyList(list);
			for(let stylename of Object.keys(Game._computers)) {
				if (!Game._orders[stylename]) {
					const term = document.createElement("dt");
					const def = document.createElement("dd");
					Util.replaceText(term,stylename);
					Util.replaceText(def,"");
					list.append(term,def);
				}
			}
		}
	},
	_processIfEverybodyReady: function() {
		let ready = true;
		for(let i=0; i<Game._players.length; i++) {
			if (!Game._orders[Game._players[i].stylename]) {
				ready = false; break;
			}
		}
		if (ready) {
			Game._processTurn();
			return true;
		} else {
			return false;
		}
	},
	_processTurn: function() {
		document.getElementById("awaitingOrders").classList.add("hidden");
		const summary = document.getElementById("summary");
		const proc = document.getElementById("processOrder");

		const updates = TurnProcessor.processTurn();

		// Let the human know of any interesting events (if any)
		if (updates.length > 0) {
			Game._showMap(Game._players[0].map); // To show the ship discoveries
			summary.classList.add("hidden");
			proc.classList.remove("hidden");
			const list = document.getElementById("messages");
			list.replaceChildren();
			for(let update of updates) {
				const element = document.createElement("li");
				element.append(update);
				list.append(element);
			}
		} else {
			Game._completeTurn(); // Just complete the turn
		}
	},
	_completeTurn: function() {
		let summary = document.getElementById("summary");
		let proc = document.getElementById("processOrder");
		proc.classList.add("hidden");
		summary.classList.remove("hidden");
		Game._newTurn();
	},
	_showLoadButton: function() {
		if (localStorage.domainOfTheAncientsSaveDate) {
			const saved = new Date(localStorage.domainOfTheAncientsSaveDate);
			const nextUpdate = Util.next_time_diff(saved);
			if (localStorage.domainOfTheAncientsTurnNumber) Util.replaceText("savegameturn",localStorage.domainOfTheAncientsTurnNumber);
			Util.replaceText("savegamedate",Util.time_diff(saved));
			// Is the welcome screen hidden?
			const hiddenWelcome = document.getElementById("welcome").classList.contains("hidden");
			if (nextUpdate && !hiddenWelcome) {
				window.setTimeout(Util.captureExceptionFunction(Game._showLoadButton, nextUpdate*1000));
			}
		}
	},
	_saveGameState: function() {
		Game._players[0].ROTRNGState = JSON.stringify(ROT.RNG.getState()); // Save the ROT state for the main thread (we use it for combats)
		let data = {};
		data.turn = Game.State.turn;
		data.baseMap = Game.State.baseMap;
		data.playerTech = Game._players[0].techIndex;
		data.players = [];
		for(let p of Game._players) {
			data.players.push(p);
		}
		data.built = [];
		for(let design of Game._standardShipDesigns) {
			data.built.push(design._built);
		}
		localStorage.domainOfTheAncientsTurnNumber = Game.State.turn;
		localStorage.domainOfTheAncientsSaveDate = (new Date()).toISOString();
		let version = {game: Game.State.version, plugin:"3.178.0"};
		localStorage.domainOfTheAncientsVersion = JSON.stringify(version);
		localStorage.domainOfTheAncients = JSON.stringify(data);
	},
	_load: function() {
		const data = JSON.parse(localStorage.domainOfTheAncients);
		Game.State.turn = data.turn;
		Game.State.baseMap = new Map(data.baseMap);
		Game._createPlayers(data.playerTech);
		for(let i=0; i<Game._players.length; i++) {
			Game._players[i].fromJSON(data.players[i],Game.State.baseMap);
			Game._stylenameToPlayerMapping[Game._players[i].stylename] = Game._players[i];
		}

		ROT.RNG.setState(JSON.parse(data.players[0].ROT));

		let i=0;
		for(let design of Game._standardShipDesigns) {
			design._built = data.built[i++];
		}

		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);

		const v = JSON.parse(localStorage.domainOfTheAncientsVersion);
		Game.State.version = v.game;

		Game.History = JSON.parse(localStorage.domainOfTheAncientsHistory);

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		Game._startTurn();
	},
};

window.onload = function() {
	Util.captureException(Game.init);
};
