/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals Util, TECHS, ShipOrder, RepairShipOrder, ScrapShipOrder, ShipDesignType, ShipDesigner */
/* exported ShipOrderReport */
let ShipOrderReport = {
	showShipReports: function(player) {
		if (player.researchLevel(TECHS.planetaryDefences) >= 3) {
			document.getElementById("andBases").classList.remove("hidden");
		}
		let ships = document.getElementById("ships");
		for(let ship of player.ships) {
			const newrow = ships.insertRow();
			const namecell = document.createElement("td"); namecell.append(ship.name);
			namecell.append(document.createElement("br"));
			const designbutton = document.createElement("button");
			designbutton.append("Review design");
			const selfDesign = ship.design;
			designbutton.onclick = Util.captureExceptionFunction(ShipDesigner.review,selfDesign);
			namecell.append(designbutton);
			const loccell = document.createElement("td"); loccell.append(ship.location);
			const repcell = document.createElement("td"); repcell.append(ship.report ? ship.report : "Nothing to report");
			if (ship.getColonyImprovement()) {
				repcell.append(document.createElement("br")); repcell.append(`Can improve a colony's class by ${ship.getColonyImprovement()}`);
			}
			if (ship.getMaxMarines()) {
				repcell.append(document.createElement("br")); repcell.append(`${ship.marines} of ${ship.getMaxMarines()} marine${ship.getMaxMarines()==1?"":"s"} on board`);
			}
			if (ship.getRepairCost()) {
				repcell.append(document.createElement("br"));
				const bold = document.createElement("b"); bold.append("Blocks needing repair: ");
				const damaged = ship.getDestroyedBlocks();
				let previous = damaged[0].name, count = 1;
				let comma = "";
				for(let i=1; i<damaged.length; i++) {
					if (previous == damaged[i].name) {
						count++;
					} else {
						if (count==1) {
							bold.append(comma+previous);
						} else {
							bold.append(comma+count+"x"+previous);
						}
						comma = ", ";
						previous = damaged[i].name;
						count = 1;
					}
				}
				if (count==1) {
					bold.append(comma+previous);
				} else {
					bold.append(comma+count+"x"+previous);
				}
				repcell.append(bold);
			}
			ship.order = new ShipOrder(ship.retreatingTo);
			if (ship.retreatingTo) ship.order.retreating = true;
			const ordercell = document.createElement("td"); ordercell.append("Hold position");
			const changecell = document.createElement("td");
			const which = {player: player, ship: ship, ordercell: ordercell, target: null, coloniseTickBox: null, embarkTickBox: null, disembarkTickBox: null };
			if (ship.getJumpRange() > 0) {
				const changeJumpButton = document.createElement("button"); changeJumpButton.append("Change Jump Target...");
				changeJumpButton.classList.add("map-using");
				if (ship.retreatingTo) {
					changeJumpButton.disabled = true;
				} else {
					changeJumpButton.classList.add("map-using");
				}
				changeJumpButton.onclick = Util.captureExceptionFunction(ShipOrderReport._changeJump,which);
				changecell.append(changeJumpButton);
				if (ship.getColonyImprovement()) {
					const coloniseTickBox = document.createElement("input"); coloniseTickBox.type = "checkbox";
					coloniseTickBox.checked = ship.order.colonise || false;
					which.coloniseTickBox = coloniseTickBox;
					coloniseTickBox.onchange = Util.captureExceptionFunction(ShipOrderReport._coloniseChange,which);
					changecell.append(document.createElement("br"));
					changecell.append(coloniseTickBox);
					changecell.append("Colonise on arrival");
				}
				if (ship.marines < ship.getMaxMarines()) {
					const count = ship.getMaxMarines() - ship.marines;
					const embarkTickBox = document.createElement("input"); embarkTickBox.type = "checkbox";
					embarkTickBox.checked = ship.order.embark || false;
					which.embarkTickBox = embarkTickBox;
					embarkTickBox.onchange = Util.captureExceptionFunction(ShipOrderReport._embarkChange,which);
					changecell.append(document.createElement("br"));
					changecell.append(embarkTickBox);
					changecell.append(`Embark ${count} marine compan${count==1?"y":"ies"}`);
				}
				if (ship.marines > 0) {
					const disembarkTickBox = document.createElement("input"); disembarkTickBox.type = "checkbox";
					disembarkTickBox.checked = ship.order.disembark || false;
					which.disembarkTickBox = disembarkTickBox;
					disembarkTickBox.onchange = Util.captureExceptionFunction(ShipOrderReport._disembarkChange,which);
					changecell.append(document.createElement("br"));
					changecell.append(disembarkTickBox);
					changecell.append("Disembark marines on arrival");
				}
			}

			const hex = player.map.hex(ship.location);
			if (ship.getRepairCost() > 0) {
			 	if (ship.retreatingTo == null && hex.planet && hex.owner == player.stylename && (ship.design.type !== ShipDesignType.Ship || hex.level > 19)) {
					const repairShipButton = document.createElement("button");
					repairShipButton.append("Repair Ship at "+ship.location+" for "+ship.getRepairCost()+" PPs");
					repairShipButton.onclick =  Util.captureExceptionFunction(ShipOrderReport._repair,which);
					changecell.append(document.createElement("br"));
					changecell.append(repairShipButton);
				} else {
					repcell.append(document.createElement("br"));
					if (ship.retreatingTo == null) {
						repcell.append("(Note: the ship cannot be repaired at this location)");
					} else {
						repcell.append("(Note: the ship cannot be repaired while it is retreating)");
					}
				}
			}

			const scrap = ship.getScrapValue();
			const scrapShipButton = document.createElement("button"); scrapShipButton.append(`Scrap ship for ${scrap} PPs`);
			scrapShipButton.onclick = Util.captureExceptionFunction(ShipOrderReport._scrap,which);
			changecell.append(document.createElement("br"));
			changecell.append(scrapShipButton);
			newrow.append(namecell,loccell,repcell,ordercell,changecell);
		}
	},
	_changeJump: function(which) {
		Util.disableMapUsingButtons(true);
		Util.replaceText("shipinstructionsShipName",which.ship.name);
		document.getElementById("shipinstructions").classList.remove("hidden");
		document.getElementById("cancelJumpChange").onclick = Util.captureExceptionFunction(ShipOrderReport._cancelJumpTarget,which);

		const location = which.ship.location;
		const jumpRange = which.ship.getJumpRange();
		which.player.computeCommsGrid();
		for(let row=1; row<21; row++) {
			for(let column of "ABCDEFGHIJKLMNOP") {
				const name = column+row;
				const hexelement = document.getElementById(name);
				if (Map.range(location,name) <= jumpRange && which.player.map.hex(name).onGrid) {
					const args = Object.assign({},which);
					args.target = name;
					hexelement.onclick = Util.captureExceptionFunction(ShipOrderReport._selectJumpTarget,args);
				} else {
					hexelement.classList.add("greyout");
				}
			}
		}
	},
	_repair: function(which) {
		which.ship.order = new RepairShipOrder(which.ship.getRepairCost());
		ShipOrderReport._fillInOrders(which);
	},
	_scrap: function(which) {
		which.ship.order = new ScrapShipOrder(which.ship.getScrapValue());
		ShipOrderReport._fillInOrders(which);
	},
	_selectJumpTarget: function(which) {
		which.ship.order = new ShipOrder(which.target,(which.coloniseTickBox && which.coloniseTickBox.checked));
		if (which.embarkTickBox && which.embarkTickBox.checked) { which.ship.order.setMarinesToEmbark(); }
		if (which.disembarkTickBox && which.disembarkTickBox.checked) { which.ship.order.setMarinesToDisembark(); }
		ShipOrderReport._finishJumpTarget(which);
	},
	_coloniseChange: function(which) {
		which.ship.order = new ShipOrder(which.ship.order.jump,(which.coloniseTickBox && which.coloniseTickBox.checked));
		if (which.embarkTickBox && which.embarkTickBox && which.embarkTickBox.checked) { which.ship.order.setMarinesToEmbark(); }
		if (which.disembarkTickBox && which.disembarkTickBox.checked) { which.ship.order.setMarinesToDisembark(); }
		ShipOrderReport._fillInOrders(which);
	},
	_embarkChange: function(which) {
		which.ship.order = new ShipOrder(which.ship.order.jump,(which.coloniseTickBox && which.coloniseTickBox.checked));
		if (which.embarkTickBox.checked) {
			which.ship.order.setMarinesToEmbark();
			if (which.disembarkTickBox) which.disembarkTickBox.checked = false;
		} else {
			if (which.disembarkTickBox.checked) { which.ship.order.setMarinesToDisembark(); }
		}
		ShipOrderReport._fillInOrders(which);
	},
	_disembarkChange: function(which) {
		which.ship.order = new ShipOrder(which.ship.order.jump,(which.coloniseTickBox && which.coloniseTickBox.checked));
		if (which.disembarkTickBox.checked) {
			which.ship.order.setMarinesToDisembark();
			if (which.embarkTickBox) which.embarkTickBox.checked = false;
		} else {
			if (which.embarkTickBox && which.embarkTickBox.checked) { which.ship.order.setMarinesToEmbark(); }
		}
		ShipOrderReport._fillInOrders(which);
	},
	_fillInOrders: function(which) {
		which.ordercell.replaceChildren();
		which.ordercell.append(which.ship.order.toString());
	},
	_cancelJumpTarget: function(which) {
		which.ship.order.cancelJump();
		ShipOrderReport._finishJumpTarget(which);
	},
	_finishJumpTarget: function(which) {
		ShipOrderReport._fillInOrders(which);
		document.getElementById("shipinstructions").classList.add("hidden");
		const cancel = document.getElementById("cancelJumpChange");
		cancel.onclick = null;
		Util.disableMapUsingButtons(false);
		for(let row=1; row<21; row++) {
			for(let column of "ABCDEFGHIJKLMNOP") {
				const name = column+row;
				const hexelement = document.getElementById(name);
				hexelement.onclick = null;
				hexelement.classList.remove("greyout");
			}
		}
	}

};